@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -243,12 +243,10 @@ update_selected (AdwTab *self)
   update_indicator (self);
 }
 
-static gboolean
+static void
 close_idle_cb (AdwTab *self)
 {
   adw_tab_view_close_page (self->view, self->page);
-
-  return G_SOURCE_REMOVE;
 }
 
 static void
@@ -260,7 +258,7 @@ close_clicked_cb (AdwTab *self)
   /* When animations are disabled, we don't want to immediately remove the
    * whole tab mid-click. Instead, defer it until the click has happened.
    */
-  g_idle_add ((GSourceFunc) close_idle_cb, self);
+  g_idle_add_once ((GSourceOnceFunc) close_idle_cb, self);
 }
 
 static void
@@ -332,8 +330,10 @@ drop_cb (AdwTab *self,
          GValue *value)
 {
   gboolean ret = GDK_EVENT_PROPAGATE;
+  GdkDrop *drop = gtk_drop_target_get_current_drop (self->drop_target);
+  GdkDragAction preferred_action = gdk_drop_get_actions (drop);
 
-  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, &ret);
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, preferred_action, &ret);
 
   return ret;
 }
@@ -603,9 +603,8 @@ adw_tab_snapshot (GtkWidget   *widget,
     gboolean is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
     int width = gtk_widget_get_width (widget);
     int height = gtk_widget_get_height (widget);
-    float offset =
-      gtk_widget_get_allocated_width (self->close_btn) +
-      gtk_widget_get_margin_end (self->title);
+    float offset = gtk_widget_get_width (self->close_btn) +
+                   gtk_widget_get_margin_end (self->title);
 
     gtk_snapshot_push_mask (snapshot, GSK_MASK_MODE_INVERTED_ALPHA);
 
@@ -800,8 +799,9 @@ adw_tab_class_init (AdwTabClass *klass)
                   g_signal_accumulator_first_wins,
                   NULL, NULL,
                   G_TYPE_BOOLEAN,
-                  1,
-                  G_TYPE_VALUE);
+                  2,
+                  G_TYPE_VALUE,
+                  GDK_TYPE_DRAG_ACTION);
 
   signals[SIGNAL_EXTRA_DRAG_VALUE] =
     g_signal_new ("extra-drag-value",
